from enum import Enum
import re
from typing import Any, Generator, TypeVar, Generic, List, Type

from iamcore.irn import IRN

from iamcore.client.exceptions import IAMException


def to_snake_case(field_name: str) -> str:
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', field_name).lower()


def to_dict(obj):
    if isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = to_dict(v)
        return data
    elif isinstance(obj, IRN):
        return str(obj)
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [to_dict(v) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = dict([(key, to_dict(value))
                     for key, value in obj.__dict__.items()
                     if not callable(value) and not key.startswith('_')])
        return data
    else:
        return obj


def generic_search_all(auth_headers: dict[str, str], func, *vargs, **kwargs) -> Generator[Any, None, None]:
    if "page" in kwargs.keys():
        kwargs.pop("page")
    new_results = True
    page = 1
    cnt = 0
    while new_results:
        resp = func(auth_headers, *vargs, page=page, **kwargs)
        if not resp.data:
            break
        for d in resp.data:
            yield d
            cnt += 1
        new_results = cnt < resp.count
        page += 1


class SortOrder(Enum):
    asc = 1
    desc = 2


T = TypeVar('T')


class IamEntityResponse(Generic[T]):
    data: T

    def __init__(self, base_class: Type[T], data: dict):
        if isinstance(data, base_class):
            self.data = data
        elif isinstance(data, dict):
            self.data = base_class(**data)


class IamEntitiesResponse(Generic[T]):
    data: List[T]
    count: int
    page: int
    page_size: int

    def __init__(self, base_class: Type[T], data: List[dict], **kwargs):
        if not isinstance(data, list):
            raise IAMException(f"Unexpected response format")

        self.data = [
            base_class.of(item)
            for item in data
        ]

        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)



from typing import List, Generator

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.common import to_snake_case, to_dict, SortOrder, generic_search_all, IamEntityResponse, \
    IamEntitiesResponse
from iamcore.client.config import config
from iamcore.client.exceptions import err_chain, IAMException, unwrap_post, unwrap_get


class ApplicationResourceType(object):
    id: str
    irn: IRN
    type: str
    description: str
    action_prefix: str
    operations: List[str]
    created: str
    updated: str

    def __init__(self, irn: str, **kwargs):
        self.irn = IRN.from_irn_str(irn)
        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)

    @staticmethod
    def of(item):
        if isinstance(item, ApplicationResourceType):
            return item
        elif isinstance(item, dict):
            return ApplicationResourceType(**item)
        raise IAMException(f"Unexpected response format")

    def to_dict(self):
        return to_dict(self)


@err_chain(IAMException)
def create_resource_type(
        auth_headers: dict[str, str],
        application_irn: IRN,
        payload: dict[str, str] = None,
        type: str = None,
        description: str = None,
        action_prefix: str = None,
        operations: List[str] = None
) -> ApplicationResourceType:
    url = config.IAMCORE_URL + "/api/v1/applications/" + IRN.of(application_irn).to_base64() + "/resource-types"
    if not payload:
        payload = {
            "type": type,
            "description": description,
            "actionPrefix": action_prefix,
            "operations": operations,
        }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    return IamEntityResponse(ApplicationResourceType, **unwrap_post(response)).data


@err_chain(IAMException)
def get_resource_type(
        auth_headers: dict[str, str],
        application_irn: IRN,
        irn: IRN,
) -> ApplicationResourceType:
    url = (config.IAMCORE_URL + "/api/v1/applications/" + IRN.of(application_irn).to_base64() +
           "/resource-types/" + IRN.of(irn).to_base64())
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("GET", url, headers=headers)
    return IamEntityResponse(ApplicationResourceType, **unwrap_get(response)).data


@err_chain(IAMException)
def search_application_resource_types(
        headers: dict[str, str],
        application_irn: IRN,
        page: int = None,
        page_size: int = None,
        sort: str = None,
        sort_order: SortOrder = None
) -> IamEntitiesResponse[ApplicationResourceType]:
    url = config.IAMCORE_URL + "/api/v1/applications/" + IRN.of(application_irn).to_base64() + "/resource-types"

    querystring = {
        "page": page,
        "pageSize": page_size,
        "sort": sort,
        "sortOrder": sort_order
    }

    querystring = {
        k: v
        for k, v in querystring.items()
        if v
    }

    response: Response = requests.request("GET", url, data="", headers=headers, params=querystring)
    return IamEntitiesResponse(ApplicationResourceType, **unwrap_get(response))


@err_chain(IAMException)
def search_all_application_resource_types(
        auth_headers: dict[str, str], *vargs, **kwargs
) -> Generator[ApplicationResourceType, None, None]:
    return generic_search_all(auth_headers, search_application_resource_types, *vargs, **kwargs)



import http.client
from typing import List, Dict, Generator

import requests
from iamcore.irn import IRN
from requests import Response

from iamcore.client.common import to_snake_case, SortOrder, to_dict, generic_search_all, IamEntityResponse, \
    IamEntitiesResponse
from iamcore.client.config import config
from iamcore.client.exceptions import IAMUnauthorizedException, err_chain, IAMResourceException, unwrap_post, \
    unwrap_put, unwrap_delete, unwrap_get, IAMException


class Resource(object):
    id: str
    irn: IRN
    name: str
    display_name: str
    description: str
    path: str
    tenant_id: str
    application: str
    resourceType: str
    enabled: bool
    metadata: Dict[str, str]
    created: str
    updated: str

    @staticmethod
    def of(item):
        if isinstance(item, Resource):
            return item
        elif isinstance(item, dict):
            return Resource(**item)
        raise IAMResourceException(f"Unexpected response format")

    def __init__(self, irn: str, **kwargs):
        self._irn = IRN.from_irn_str(irn)
        for k, v in kwargs.items():
            attr = to_snake_case(k)
            setattr(self, attr, v)

    def delete(self, auth_headers: Dict[str, str]):
        delete_resource(auth_headers, self.id)

    def update(self, auth_headers: Dict[str, str]):
        update_resource(
            auth_headers,
            resource_id=self.id,
            display_name=self.display_name,
            enabled=self.enabled,
            description=self.description,
            metadata=self.metadata
        )

    def to_dict(self):
        return to_dict(self)


@err_chain(IAMResourceException)
def create_resource(
        auth_headers: dict[str, str],
        payload: dict[str, object] = None,
        name: str = None,
        display_name: str = None,
        tenant_id: str = None,
        application: str = None,
        path: str = None,
        resource_type: str = None,
        enabled: bool = True,
        description: str = None,
        metadata: Dict[str, object] = None
) -> Resource:
    url = config.IAMCORE_URL + "/api/v1/resources"
    if not payload:
        payload = {
            "name": name,
            "displayName": display_name,
            "tenantID": tenant_id,
            "application": application,
            "path": path,
            "resourceType": resource_type,
            "enabled": enabled,
            "description": description,
            "metadata": metadata
        }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    return IamEntityResponse(Resource, **unwrap_post(response)).data


@err_chain(IAMResourceException)
def update_resource(
        auth_headers: dict[str, str],
        payload: dict[str, object] = None,
        resource_id: str = None,
        display_name: str = None,
        enabled: bool = True,
        description: str = None,
        metadata: Dict[str, object] = None
) -> None:
    if not auth_headers:
        raise IAMUnauthorizedException(f"Missing authorization headers")
    if not resource_id:
        raise IAMResourceException(f"Missing resource_id")
    url = config.IAMCORE_URL + "/api/v1/resources/" + IRN.of(resource_id).to_base64()
    if not payload:
        payload = {
            "displayName": display_name,
            "enabled": enabled,
            "description": description,
            "metadata": metadata
        }
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("PATCH", url, json=payload, headers=headers)
    unwrap_put(response)


@err_chain(IAMResourceException)
def delete_resource(auth_headers: dict[str, str], resource_id: str) -> None:
    if not auth_headers:
        raise IAMUnauthorizedException(f"Missing authorization headers")
    if not resource_id:
        raise IAMResourceException(f"Missing resource_id")

    url = config.IAMCORE_URL + "/api/v1/resources/" + IRN.of(resource_id).to_base64()
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    response: Response = requests.request("DELETE", url, data="", headers=headers)
    unwrap_delete(response)


@err_chain(IAMResourceException)
def delete_resources(auth_headers: dict[str, str], resources_ids: List[IRN]) -> None:
    if not auth_headers:
        raise IAMUnauthorizedException(f"Missing authorization headers")
    if not resources_ids:
        raise IAMResourceException(f"Missing resource_id")

    url = config.IAMCORE_URL + "/api/v1/resources/delete"
    headers = {
        "Content-Type": "application/json",
        **auth_headers
    }
    payload = {
        "resourceIDs": [
            IRN.of(r).to_base64()
            for r in resources_ids
            if r
        ]
    }
    response: Response = requests.request("POST", url, json=payload, headers=headers)
    unwrap_delete(response)


@err_chain(IAMResourceException)
def search_resource(
        headers: dict[str, str],
        irn: IRN = None,
        path: str = None,
        display_name: str = None,
        enabled: bool = None,
        tenant_id: str = None,
        application: str = None,
        resource_type: str = None,
        page: int = None,
        page_size: int = None,
        sort: str = None,
        sort_order: SortOrder = None
) -> IamEntitiesResponse[Resource]:
    url = config.IAMCORE_URL + "/api/v1/resources"

    querystring = {
        "irn": str(irn) if irn else None,
        "path": path,
        "application": application,
        "enabled": enabled,
        "displayName": display_name,
        "resourceType": resource_type,
        "tenantID": tenant_id,
        "page": page,
        "pageSize": page_size,
        "sort": sort,
        "sortOrder": sort_order
    }

    querystring = {
        k: v
        for k, v in querystring.items()
        if v
    }

    response = requests.request("GET", url, data="", headers=headers, params=querystring)
    return IamEntitiesResponse(Resource, **unwrap_get(response))


@err_chain(IAMException)
def search_all_resources(auth_headers: dict[str, str], *vargs, **kwargs) -> Generator[Resource, None, None]:
    return generic_search_all(auth_headers, search_resource, *vargs, **kwargs)



import unittest
import pytest
from iamcore.irn import IRN

from iamcore.client.auth import get_token_with_password, TokenResponse
from iamcore.client.tenant import search_tenant, create_tenant
from iamcore.client.config import config
from iamcore.client.policy import search_policy, CreatePolicyRequest, search_all_policies
from tests.conf import IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD


@pytest.fixture(scope="class")
def root_token(request):
    request.cls.root = get_token_with_password("root", config.SYSTEM_BACKEND_CLIENT_ID,
                                               IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD)


@pytest.fixture(scope="class")
def test_tenant(request):
    request.cls.tenant_name = "iamcore-py-test-policy-tenant"
    request.cls.tenant_display_name = "iamcore_ Python Sdk test policy tenant"
    request.cls.policy_name = "allow-all-iamcore-py-test-policy-CrudPoliciesTestCase"
    request.cls.policy_name_account = request.cls.policy_name + "-account-lvl"
    request.cls.policy_name_tenant = request.cls.policy_name + "-tenant-lvl"
    request.cls.policy_description = "Allow all for iamcore-py-test-policy-tenant tenant"


@pytest.mark.usefixtures("root_token")
@pytest.mark.usefixtures("test_tenant")
class CrudPoliciesTestCase(unittest.TestCase):
    root: TokenResponse
    tenant_name: str
    tenant_display_name: str
    policy_name: str
    policy_name_account: str
    policy_name_tenant: str
    policy_description: str

    def test_00_cleanup_ok(self):
        policies = search_policy(self.root.access_headers, name=self.policy_name).data
        if policies:
            self.assertEqual(len(policies), 0)
            for policy in policies:
                self.assertEqual(policy.name, self.policy_name)
                self.assertTrue(policy.id)
                self.assertTrue(policy.irn)
                self.assertTrue(policy.description)
                policy.delete(self.root.access_headers)
        policies = search_policy(self.root.access_headers, name=self.policy_name)
        self.assertEqual(len(policies.data), 0)

    def test_10_account_policy_ok(self):
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        tenant = tenants[0] if len(tenants) > 0 else \
            create_tenant(self.root.access_headers, name=self.tenant_name, display_name=self.tenant_display_name)

        policy_req = CreatePolicyRequest(self.policy_name_account, 'account', self.policy_description)
        account = IRN.of(tenant.irn).account_id
        policy_req \
            .with_statement('allow', self.policy_description, [f"irn:{account}:unittest:{tenant.tenant_id}:*"], ['*']) \
            .create(self.root.access_headers)

        policies = search_policy(self.root.access_headers, name=self.policy_name_account).data
        if policies:
            self.assertEqual(len(policies), 1)
            created_policy = policies[0]

            self.assertEqual(created_policy.name, self.policy_name_account)
            self.assertEqual(created_policy.description, self.policy_description)
            self.assertTrue(created_policy.id)
            self.assertTrue(created_policy.irn)
            # assert that we create an account policy
            self.assertEqual(IRN.of(created_policy.irn).tenant_id, '')
            self.assertTrue(created_policy.statements)

    def test_10_tenant_policy_ok(self):
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        tenant = tenants[0] if len(tenants) > 0 else \
            create_tenant(self.root.access_headers, name=self.tenant_name, display_name=self.tenant_display_name)

        policy_req = CreatePolicyRequest(self.policy_name_tenant, 'tenant', self.policy_description,
                                         tenant_id=tenant.tenant_id)
        account = IRN.of(tenant.irn).account_id
        policy_req \
            .with_statement('allow', self.policy_description, [f"irn:{account}:unittest:{tenant.tenant_id}:*"], ['*']) \
            .create(self.root.access_headers)

        policies = list(search_all_policies(self.root.access_headers, name=self.policy_name_tenant, account_id=account,
                                            tenant_id="wrongId"))
        self.assertEqual(len(policies), 0)

        policies = search_policy(self.root.access_headers, name=self.policy_name_tenant, account_id=account,
                                 tenant_id=tenant.tenant_id).data
        if policies:
            self.assertEqual(len(policies), 1)
            created_policy = policies[0]

            self.assertEqual(created_policy.name, self.policy_name_tenant)
            self.assertEqual(created_policy.description, self.policy_description)
            self.assertTrue(created_policy.id)
            self.assertTrue(created_policy.irn)
            # assert that we create an account policy
            self.assertEqual(IRN.of(created_policy.irn).tenant_id, tenant.tenant_id)
            self.assertTrue(created_policy.statements)

    def test_90_cleanup_ok(self):
        policies = search_policy(self.root.access_headers, name=self.policy_name).data
        if policies:
            self.assertEqual(len(policies), 2)
            for policy in policies:
                self.assertIn(policy.name, (self.policy_name_tenant, self.policy_name_account))
                self.assertTrue(policy.id)
                self.assertTrue(policy.irn)
                self.assertTrue(policy.description)
                policy.delete(self.root.access_headers)
        policies = search_policy(self.root.access_headers, name=self.policy_name).data
        self.assertEqual(len(policies), 0)



import unittest
import pytest
from iamcore.irn import IRN

from iamcore.client.auth import get_token_with_password, TokenResponse
from iamcore.client.tenant import search_tenant, create_tenant, get_issuer
from iamcore.client.config import config

from tests.conf import IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD


@pytest.fixture(scope="class")
def root_token(request):
    request.cls.root = get_token_with_password("root", config.SYSTEM_BACKEND_CLIENT_ID,
                                               IAMCORE_ROOT_USER, IAMCORE_ROOT_PASSWORD)


@pytest.fixture(scope="class")
def test_tenant(request):
    request.cls.tenant_name = "iamcore-py-test-tenant"
    request.cls.tenant_display_name = "iamcore_ Python Sdk test tenant"
    request.cls.login_theme = 'kaa'


@pytest.mark.usefixtures("root_token")
@pytest.mark.usefixtures("test_tenant")
class CrudTenantsTestCase(unittest.TestCase):
    root: TokenResponse
    tenant_name: str
    tenant_display_name: str
    login_theme: str

    def test_00_cleanup_ok(self):
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        if tenants:
            self.assertLessEqual(len(tenants), 1)
            for tenant in tenants:
                self.assertEqual(tenant.name, self.tenant_name)
                self.assertTrue(tenant.display_name)
                self.assertTrue(tenant.irn)
                self.assertTrue(tenant.tenant_id)
                self.assertTrue(tenant.resource_id)
                self.assertTrue(tenant.created)
                self.assertTrue(tenant.updated)
                tenant.delete(self.root.access_headers)
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        self.assertFalse(tenants)

    def test_10_crud_ok(self):
        name = self.tenant_name + '_10'
        tenant_display_name = self.tenant_display_name + '_10'
        tenant = create_tenant(self.root.access_headers, name=name, display_name=tenant_display_name,
                               login_theme=self.login_theme)

        self.assertEqual(tenant.name, name)
        self.assertEqual(tenant.display_name, tenant_display_name)
        self.assertTrue(tenant.irn)
        self.assertTrue(tenant.tenant_id)
        self.assertTrue(tenant.resource_id)
        self.assertTrue(tenant.created)
        self.assertTrue(tenant.updated)
        self.assertEqual(self.login_theme, tenant.login_theme)

    def test_12_default_login_theme_ok(self):
        name = self.tenant_name + '_12'
        tenant_display_name = self.tenant_display_name + '_12'
        tenant = create_tenant(self.root.access_headers, name=name, display_name=tenant_display_name)

        self.assertEqual(tenant.name, name)
        self.assertEqual(tenant.display_name, tenant_display_name)
        self.assertTrue(tenant.irn)
        self.assertTrue(tenant.tenant_id)
        self.assertTrue(tenant.resource_id)
        self.assertTrue(tenant.created)
        self.assertTrue(tenant.updated)
        self.assertEqual('iamcore', tenant.login_theme)

    def test_14_update_ok(self):
        name = self.tenant_name + '_12'
        tenant_display_name = self.tenant_display_name + '_12'
        tenants = search_tenant(self.root.access_headers, name=name).data
        self.assertEqual(len(tenants), 1)
        tenant = tenants[0]

        tenant.display_name = tenant_display_name + " updated"

        tenant.update(self.root.access_headers)
        self.assertEqual(tenant.name, name)
        self.assertEqual(tenant.display_name, tenant_display_name + " updated")
        self.assertTrue(tenant.irn)
        self.assertTrue(tenant.tenant_id)
        self.assertTrue(tenant.resource_id)
        self.assertTrue(tenant.created)
        self.assertTrue(tenant.updated)

    def test_20_issuer_ok(self):
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name+'_10').data
        self.assertLessEqual(len(tenants), 1)
        tenant = tenants[0]
        account = IRN.of(tenant.irn).account_id
        issuer = get_issuer(self.root.access_headers, account, tenant.tenant_id)
        self.assertTrue(issuer)
        self.assertEqual(str(issuer.irn), str(IRN.of(f'irn:{account}:iamcore:{tenant.tenant_id}::issuer/iamcore')))
        self.assertEqual(issuer.url, f"{config.IAMCORE_ISSUER_URL.strip()}/realms/{tenant.tenant_id}")
        self.assertEqual(issuer.login_url,
                         f"{config.IAMCORE_ISSUER_URL.strip()}/realms/{tenant.tenant_id}/protocol/openid-connect/auth")
        self.assertTrue(issuer.client_id)

    def test_20_search_ok(self):
        name = self.tenant_name + '_10'
        tenant_display_name = self.tenant_display_name + '_10'
        tenants = search_tenant(self.root.access_headers, name=name).data
        self.assertEqual(len(tenants), 1)
        tenant = tenants[0]

        search_plan = [
            ('name', tenant.name),
            ('display_name', tenant.display_name),
            ('irn', tenant.irn),
            ('tenant_id', tenant.tenant_id)
        ]

        for param, value in search_plan:
            tenants = search_tenant(self.root.access_headers, **{param: value}).data
            self.assertEqual(len(tenants), 1)
            self.assertTrue(tenants[0].name.startswith(tenant.name))
            self.assertEqual(tenants[0].display_name, tenant.display_name)
            self.assertTrue(tenants[0].irn, tenant.irn)
            self.assertTrue(tenants[0].tenant_id, tenant.tenant_id)
            self.assertTrue(tenants[0].resource_id, tenant.resource_id)
            self.assertTrue(tenants[0].created, tenant.created)
            self.assertTrue(tenants[0].updated, tenant.updated)

    def test_90_cleanup_ok(self):
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        if tenants:
            self.assertLessEqual(len(tenants), 2)
            for tenant in tenants:
                self.assertTrue(tenant.name.startswith(self.tenant_name))
                self.assertTrue(tenant.display_name)
                self.assertTrue(tenant.irn)
                self.assertTrue(tenant.tenant_id)
                self.assertTrue(tenant.resource_id)
                self.assertTrue(tenant.created)
                self.assertTrue(tenant.updated)
                tenant.delete(self.root.access_headers)
        tenants = search_tenant(self.root.access_headers, name=self.tenant_name).data
        self.assertFalse(tenants)
